use enum_display::EnumDisplay;
use std::string::FromUtf8Error;
use tokio_rustls::rustls;

use log::error;
use quick_xml::events::attributes::AttrError;

#[derive(Debug, EnumDisplay, Clone, Copy)]
#[enum_display(case = "Kebab")]
#[allow(unused)]
pub enum StreamError {
    /// The entity has sent XML that cannot be processed.
    BadFormat,
    /// The entity has sent a namespace prefix that is unsupported, or has
    /// sent no namespace prefix on an element that needs such a prefix
    BadNamespacePrefix,
    /// The server either (1) is closing the existing stream for this entity
    /// because a new stream has been initiated that conflicts with the
    /// existing stream, or (2) is refusing a new stream for this entity
    /// because allowing the new stream would conflict with an existing stream
    Conflict,
    /// One party is closing the stream because it has reason to believe that
    /// the other party has permanently lost the ability to communicate over
    /// the stream
    ConnectionTimeout,
    /// The value of the 'to' attribute provided in the initial stream header
    /// corresponds to an FQDN that is no longer serviced by the receiving entity
    HostGone,
    /// The value of the 'to' attribute provided in the initial stream header
    /// does not correspond to an FQDN that is serviced by the receiving entity
    HostUnknown,
    /// A stanza sent between two servers lacks a 'to' or 'from' attribute,
    /// the 'from' or 'to' attribute has no value, or the value violates the
    /// rules for XMPP addresses
    ImproperAddressing,
    /// The server has experienced a misconfiguration or other internal error
    /// that prevents it from servicing the stream
    InternalServerError,
    /// The data provided in a 'from' attribute does not match an authorized
    /// JID or validated domain as negotiated (1) between two servers using
    /// SASL or Server Dialback, or (2) between a client and a server via
    /// SASL authentication and resource binding
    InvalidFrom,
    /// The stream namespace name is something other than
    /// "http://etherx.jabber.org/streams" or the content
    /// namespace declared as the default namespace is not supported
    /// (e.g., something other than "jabber:client" or "jabber:server").
    InvalidNamespace,
    /// The entity has sent invalid XML over the stream to a server that
    /// performs validation
    InvalidXml,
    /// The entity has attempted to send XML stanzas or other outbound data
    /// before the stream has been authenticated, or otherwise is not
    /// authorized to perform an action related to stream negotiation; the
    /// receiving entity MUST NOT process the offending data before sending
    /// the stream error.
    NotAuthorized,
    /// The initiating entity has sent XML that violates the well-formedness
    /// rules of [XML](http://www.w3.org/TR/2008/REC-xml-20081126) or [XML-NAMES](http://www.w3.org/TR/2008/REC-xml-20081126).
    NotWellFormed,
    /// The entity has violated some local service policy (e.g., a stanza
    /// exceeds a configured size limit); the server MAY choose to specify
    /// the policy in the <text/> element or in an application-specific
    /// condition element
    PolicyViolation,
    /// The server is unable to properly connect to a remote entity that is
    /// needed for authentication or authorization.
    /// This condition is not to be used when the cause of the error is within the
    /// administrative domain of the XMPP service provider, in which case the
    /// [`StreamError::InternalServerError`] condition is more appropriate.
    RemoteConnectionFailed,
    /// The server is closing the stream because it has new (typically security-critical)
    /// features to offer, because the keys or
    /// certificates used to establish a secure context for the stream have
    /// expired or have been revoked during the life of the stream
    /// because the TLS sequence number has wrapped, etc.
    /// The reset applies to the stream and to any
    /// security context established for that stream (e.g., via TLS and SASL),
    /// which means that encryption and authentication need to be
    /// negotiated again for the new stream (e.g., TLS session resumption cannot be used)
    Reset,
    /// The server lacks the system resources necessary to service the stream
    ResourceConstraint,
    /// The entity has attempted to send restricted XML features such as a
    /// comment, processing instruction, DTD subset, or XML entity reference
    RestrictedXml,
    /// The server will not provide service to the initiating entity but is
    /// redirecting traffic to another host under the administrative control
    /// of the same service provider.  The XML character data of the <see-
    /// other-host/> element returned by the server MUST specify the
    /// alternate FQDN or IP address at which to connect, which MUST be a
    /// valid domainpart or a domainpart plus port number (separated by the
    /// ':' character in the form "domainpart:port").  If the domainpart is
    /// the same as the source domain, derived domain, or resolved IPv4 or
    /// IPv6 address to which the initiating entity originally connected
    /// (differing only by the port number), then the initiating entity
    /// SHOULD simply attempt to reconnect at that address.
    /// (The format of an IPv6 address MUST follow [IPv6-ADDR](https://www.rfc-editor.org/rfc/rfc5952),
    /// which includes the enclosing the IPv6 address in square brackets '[' and ']'
    /// as originally defined by [URI](https://www.rfc-editor.org/rfc/rfc3986).)
    /// Otherwise, the initiating entity MUST resolve the FQDN
    /// specified in the <see-other-host/> element
    SeeOtherHost,
    /// The server is being shut down and all active streams are being closed
    SystemShutdown,
    /// The error condition is not one of those defined by the other
    /// conditions in this list; this error condition SHOULD NOT be used
    /// except in conjunction with an application-specific condition.
    UndefinedCondition,
    /// The initiating entity has encoded the stream in an encoding that is
    /// not supported by the server (see Section 11.6) or has otherwise
    /// improperly encoded the stream (e.g., by violating the rules of the
    /// [UTF-8](https://www.rfc-editor.org/rfc/rfc3629) encoding).
    UnsupportedEncoding,
    /// The receiving entity has advertised a mandatory-to-negotiate stream
    /// feature that the initiating entity does not support, and has offered
    /// no other mandatory-to-negotiate feature alongside the unsupported feature.
    UnsupportedFeature,
    /// The initiating entity has sent a first-level child of the stream that
    /// is not supported by the server, either because the receiving entity
    /// does not understand the namespace or because the receiving entity
    /// does not understand the element name for the applicable namespace
    /// (which might be the content namespace declared as the default namespace).
    UnsupportedStanzaType,
    ///  The 'version' attribute provided by the initiating entity in the
    /// stream header specifies a version of XMPP that is not supported by
    /// the server.
    UnsupportedVersion,
}
impl From<rustls::Error> for StreamError {
    fn from(value: rustls::Error) -> Self {
        match value {
            rustls::Error::InappropriateMessage {
                expect_types,
                got_type,
            } => Self::BadFormat,
            rustls::Error::InappropriateHandshakeMessage {
                expect_types,
                got_type,
            } => Self::BadFormat,
            rustls::Error::InvalidMessage(_) => Self::BadFormat,
            rustls::Error::NoCertificatesPresented => Self::InternalServerError,
            rustls::Error::UnsupportedNameType => Self::BadFormat,
            rustls::Error::DecryptError => Self::InternalServerError,
            rustls::Error::EncryptError => Self::InternalServerError,
            rustls::Error::PeerIncompatible(_) => Self::BadFormat,
            rustls::Error::PeerMisbehaved(_) => Self::BadFormat,
            rustls::Error::AlertReceived(_) => Self::BadFormat,
            rustls::Error::InvalidCertificate(_) => Self::InternalServerError,
            _ => Self::BadFormat,
        }
    }
}

impl From<std::io::Error> for StreamError {
    fn from(value: std::io::Error) -> Self {
        error!("io error: {value}");
        Self::InternalServerError
    }
}

impl From<FromUtf8Error> for StreamError {
    fn from(_: FromUtf8Error) -> Self {
        Self::UnsupportedEncoding
    }
}

impl From<AttrError> for StreamError {
    fn from(_: AttrError) -> Self {
        Self::NotWellFormed
    }
}

impl From<quick_xml::Error> for StreamError {
    fn from(value: quick_xml::Error) -> Self {
        match value {
            quick_xml::Error::Io(err) => {
                error!("io error: {err}");
                Self::InternalServerError
            }
            quick_xml::Error::NonDecodable(_) => Self::UnsupportedEncoding,
            _ => Self::BadFormat,
        }
    }
}

impl std::error::Error for StreamError {}
