use std::{process, sync::Arc};

use log::{error, info};

mod common;
mod config;
mod error;
mod feature;
mod negotiator;
mod server;
mod streamstart;
mod tag;
mod tls;

#[tokio::main]
async fn main() -> Result<(), anyhow::Error> {
    if !std::env::var("LOG").is_ok() {
        #[cfg(debug_assertions)]
        std::env::set_var("LOG", "debug");
        #[cfg(not(debug_assertions))]
        std::env::set_var("LOG", "info");
    }
    pretty_env_logger::init_custom_env("LOG");

    let cfg = match config::Config::load() {
        Ok(cfg) => cfg,
        Err(err) => {
            error!(
                "getting config: {err}. writing default to {}",
                config::DEFAULT_PATH
            );
            config::Config::default().save(config::DEFAULT_PATH)?;
            process::exit(1);
        }
    };
    info!("checking for certificates");
    let certs = Arc::new(
        cfg.certificate()
            .await
            .expect("getting certificates")
            .into_config()?,
    );

    let host = cfg.hostname();
    info!(
        "listening on {host} {} (plain) and {} (tls)!",
        cfg.insecure_port, cfg.tls_port
    );
    let cert_clone = certs.clone();
    let host_clone = host.clone();
    tokio::spawn(async move {
        server::listen_tls(host_clone, cfg.tls_port, cert_clone)
            .await
            .expect("TLS listener")
    });
    server::listen_starttls(host, cfg.insecure_port, certs)
        .await
        .expect("STARTTLS listener");
    Ok(())
}
