use log::{error, info};
use quick_xml::{
    events::{BytesEnd, BytesStart, Event},
    Reader, Writer,
};
use tokio::{
    io::{AsyncWriteExt, BufReader, ReadHalf, WriteHalf},
    net::TcpStream,
};
use tokio_rustls::server::TlsStream;

use crate::{
    common::{self, StreamHeader},
    error::StreamError,
    feature::Feature,
    tag::{self, Tag},
};
type Result<T> = std::result::Result<T, StreamError>;
pub struct TLSStream {
    id: String,
    reader: Reader<BufReader<ReadHalf<TlsStream<TcpStream>>>>,
    writer: Writer<WriteHalf<TlsStream<TcpStream>>>,
    buffer: Vec<u8>,
    hostname: String,
    features: Vec<Feature>,
}

impl TLSStream {
    pub fn new(conn: TlsStream<TcpStream>, hostname: String) -> Self {
        let (read, write) = tokio::io::split(conn);
        let (reader, writer) = (
            Reader::from_reader(BufReader::new(read)),
            Writer::new(write),
        );

        Self {
            reader,
            writer,
            hostname,
            id: common::gen_id(),
            buffer: vec![],
            features: vec![],
        }
    }

    pub async fn start_stream(mut self) {
        if let Err(err) = self.handle_stream().await {
            if let Err(err2) = common::error(self.writer.get_mut(), err).await {
                error!("error writing error: {err2}");
                return;
            } else {
                info!("wrote error {err}")
            }

            if let Err(e) = self.writer.get_mut().write_all(b"</stream:stream>").await {
                error!("writing end to stream: {e}")
            }
            if let Err(e) = self.writer.get_mut().shutdown().await {
                error!("shutting down stream: {e}")
            }
        }
    }

    async fn handle_stream(&mut self) -> Result<()> {
        let header = self.get_stream_header().await?;
        if !header
            .to
            .unwrap_or_default()
            .eq_ignore_ascii_case(&self.hostname)
        {
            return Err(StreamError::HostUnknown);
        }
        common::write_stream_header(
            self.writer.get_mut(),
            StreamHeader {
                from: Some(self.hostname.clone()),
                to: header.from,
                id: Some(self.id.clone()),
                namespace: common::XMLNamespace::JabberClient,
            },
        )
        .await?;
        self.send_features().await?;
        self.negotiate_features().await?;

        todo!()
    }

    async fn negotiate_features(&mut self) -> Result<()> {
        let available_features: Vec<&Feature> = (&self.features).into_iter().collect();
        while available_features.len() > 0 {
            todo!()
        }
        Ok(())
    }

    async fn send_features(&mut self) -> Result<()> {
        self.writer
            .write_event_async(Event::Start(BytesStart::new(tag::FEATURE)))
            .await?;
        for feature in &self.features {
            feature.write_tag(self.writer.get_mut()).await?;
        }
        self.writer
            .write_event_async(Event::End(BytesEnd::new(tag::FEATURE)))
            .await?;
        Ok(())
    }

    async fn get_stream_header(&mut self) -> Result<StreamHeader> {
        loop {
            match self.reader.read_event_into_async(&mut self.buffer).await? {
                Event::Start(start) => {
                    if start.name().as_ref() == tag::STREAM_ELEMENT_NAME {
                        break Ok(start.attributes().try_into()?);
                    } else {
                        break Err(StreamError::BadFormat);
                    }
                }
                Event::Decl(_) => continue,
                _ => break Err(StreamError::BadFormat),
            }
        }
    }
}
